from inspect import signature
from typing import Any, Dict, List, Optional, Tuple, Type, TypeVar, Union, cast

_T = TypeVar('_T')

_primitives = (bytes, bytearray, bool, int, float, str, dict, tuple, list, set, slice, map, zip)


def _is_primitive(val: Any) -> bool:
    return isinstance(val, _primitives)  # type: ignore


def _is_object(val: Any) -> bool:
    return isinstance(val, object) and not isinstance(val, type) and not _is_primitive(val)


def _is_simple(val: Any) -> bool:
    return not _is_object(val)


ContainerKey = Union[str, Type[Any], object]


class Container(dict):
    debug: bool = False

    def resolve(self, items: List[Union[ContainerKey, Tuple[ContainerKey, _T]]]) -> None:
        items_ = list(map(self._sanitize_item_before_resolve, items))
        while items_:
            for index, item in enumerate(items_):
                # Check if already exist
                if item[0] in self or item[1] in self:
                    if self.debug:
                        print('Ignoring {} - {}'.format(item[0], item[1]))
                    del items_[index]
                    continue
                # Resolve 2nd arg if is a primitive or instance
                if not isinstance(item[1], type):
                    if self.debug:
                        print('Adding {} - {}'.format(item[0], item[1]))
                    self.set(item[0], item[1])
                    del items_[index]
                    continue
                # Resolve or Postpone 2nd arg if is a type
                kwargs = self._resolve_or_postpone_item(item, items_)
                if kwargs is not None:
                    if self.debug:
                        print('Resolving {}'.format(item[1]))
                    inst = item[1](**kwargs)
                    if self.debug:
                        print('Adding {} - {}'.format(item[0], item[1]))
                    self.set(item[0], inst)
                    del items_[index]

    def set(self, key: ContainerKey, val: _T = ...) -> None:  # type: ignore
        """
        e.g. 1
        container = Container()
        container.set('config', 'hello world')  # {'config': 'hello world'}
        container.set(MyClass, my_class)  # {'python.path.to.my_class': '...'}
        container.set(my_another_class)  # {'python.path.to.my_another_class': '...'}
        e.g. 2
        container = Container({'config': {}})
        container.set('config.version', '0.1.0')  # {'config': {'version': '0.1.0'}}
        """
        here = self
        if isinstance(key, type):
            key = f'{key.__module__}.{key.__name__}'
        if _is_object(key):
            val = key  # type: ignore
            key = f'{key.__class__.__module__}.{key.__class__.__name__}'
        keys = cast(str, key).split('.')
        for key in keys[:-1]:
            here = here.setdefault(key, {})
        here[keys[-1]] = val

    def get(self, key: ContainerKey, typ: Optional[Type[_T]] = None) -> _T:  # type: ignore
        """
        e.g. 1
        container = Container({'config': {'version': '0.1.0'}, 'app.libs.MyClass': '...'})
        container.get('config.version')  # '0.1.0'
        container.get(MyClass)  # '...'
        e.g. 2
        container = Container({'config': {'version': '0.1.0'})
        container.get('config.version', typ=str)  # Checks type
        """
        here = self
        if isinstance(key, type) and not typ:
            key = f'{key.__module__}.{key.__name__}'
        if _is_object(key) and not typ:
            key = f'{key.__class__.__module__}.{key.__class__.__name__}'
        keys = cast(str, key).split('.')
        for key in keys[:-1]:
            if key in here and isinstance(here[key], dict):
                here = here[key]
        try:
            val = here[keys[-1]]
            if typ and not isinstance(val, (typ,)):
                raise TypeError(f'<{key}{f": {typ.__name__}" if typ else ""}> does not exist in container')
            return val
        except KeyError:
            raise KeyError(f'<{key}{f": {typ.__name__}" if typ else ""}> does not exist in container')

    def __contains__(self, *o) -> bool:  # type: ignore
        """
        e.g. 1
        container = Container({'config': '...'})
        'config' in container # True
        e.g. 2
        container = Container({'config': {'version': '0.1.0'})
        'config.version' in container # True
        'config.foo' in container # False
        """
        try:
            self.get(o[0])
            return True
        except (IndexError, KeyError, TypeError):
            return False

    @staticmethod
    def _sanitize_item_before_resolve(item: Union[ContainerKey, Tuple[ContainerKey, _T]]) -> Tuple[ContainerKey, _T]:
        item_ = item
        if not isinstance(item_, tuple):
            item_ = (item_, item_)
        if len(item_) < 2:
            item_ = (item_[0], item_[0])
        return item_  # type: ignore

    def _resolve_or_postpone_item(
        self, item: Tuple[ContainerKey, _T], items: List[Tuple[ContainerKey, _T]]
    ) -> Optional[Dict[str, Any]]:
        parameters = signature(item[1]).parameters.items()  # type: ignore
        kwargs: Dict[str, Any] = {}
        for parameter in parameters:
            typ: Type[Any] = parameter[1].annotation
            if typ in _primitives:
                raise TypeError(f'Parameter {parameter[0]} can not be primitive to self-resolve')
            if typ in self:
                kwargs.update({parameter[0]: self.get(typ)})
                continue
            if self.debug:
                print('Postponing {}'.format(typ))
            items.append((typ, typ))  # type: ignore
            kwargs = {}
            break
        if len(parameters) == len(kwargs.keys()):
            return kwargs
        return None



from unittest.mock import Mock

import pytest

from aioddd import (
    Command,
    CommandNotRegisteredError,
    Query,
    QueryNotRegisteredError,
    SimpleCommandBus,
    SimpleQueryBus,
)
from aioddd.testing import AsyncMock


@pytest.mark.asyncio
async def test_simple_command_bus() -> None:
    command_handler_mock1 = Mock()
    command_handler_mock2 = Mock()
    command_handler_mock3 = Mock()

    class _CommandTest1(Command):
        pass

    class _CommandTest2(Command):
        pass

    class _CommandTest3(Command):
        pass

    command = _CommandTest1()

    command_handler_mock1.subscribed_to = lambda: _CommandTest1
    command_handler_mock1.handle = AsyncMock(return_value=None)
    command_handler_mock2.subscribed_to = lambda: _CommandTest2
    command_handler_mock2.handle = AsyncMock(return_value=None)
    command_handler_mock3.subscribed_to = lambda: _CommandTest3
    command_handler_mock3.handle = AsyncMock(return_value=None)

    bus = SimpleCommandBus(handlers=[command_handler_mock1])
    bus.add_handler(handler=command_handler_mock2)
    bus.add_handler(handler=[command_handler_mock3])

    await bus.dispatch(command=command)

    command_handler_mock1.handle.assert_called_once()
    command_handler_mock2.handle.assert_not_called()
    command_handler_mock3.handle.assert_not_called()


@pytest.mark.asyncio
async def test_simple_command_bus_fails_because_command_was_not_registered() -> None:
    class _CommandTest(Command):
        pass

    command = _CommandTest()
    bus = SimpleCommandBus(handlers=[])

    with pytest.raises(CommandNotRegisteredError):
        await bus.dispatch(command=command)


@pytest.mark.asyncio
async def test_simple_query_bus() -> None:
    query_handler_mock1 = Mock()
    query_handler_mock2 = Mock()
    query_handler_mock3 = Mock()

    class _QueryTest1(Query):
        pass

    class _QueryTest2(Query):
        pass

    class _QueryTest3(Query):
        pass

    query = _QueryTest1()

    query_handler_mock1.subscribed_to = lambda: _QueryTest1
    query_handler_mock1.handle = AsyncMock(return_value='test')
    query_handler_mock2.subscribed_to = lambda: _QueryTest2
    query_handler_mock2.handle = AsyncMock(return_value=None)
    query_handler_mock3.subscribed_to = lambda: _QueryTest3
    query_handler_mock3.handle = AsyncMock(return_value=None)

    bus = SimpleQueryBus(handlers=[query_handler_mock1])
    bus.add_handler(handler=query_handler_mock2)
    bus.add_handler(handler=[query_handler_mock3])

    res = await bus.ask(query=query)

    query_handler_mock1.handle.assert_called_once()
    query_handler_mock2.handle.assert_not_called()
    query_handler_mock3.handle.assert_not_called()
    assert res == 'test'


@pytest.mark.asyncio
async def test_simple_query_bus_fails_because_query_was_not_registered() -> None:
    class _QueryTest(Query):
        pass

    query = _QueryTest()
    bus = SimpleQueryBus(handlers=[])

    with pytest.raises(QueryNotRegisteredError):
        await bus.ask(query=query)



import sys
from datetime import datetime

if sys.version_info.major == 3 and sys.version_info.minor <= 6:
    from backports.datetime_fromisoformat import MonkeyPatch

    MonkeyPatch.patch_fromisoformat()


datetime_fromisoformat = getattr(datetime, 'fromisoformat')



from dataclasses import asdict, dataclass
from unittest.mock import Mock

import pytest

from aioddd import (
    ConfigEventMappers,
    Event,
    EventBus,
    EventMapper,
    EventMapperNotFoundError,
    EventPublisher,
    EventPublishers,
    Id,
    InternalEventPublisher,
    SimpleEventBus,
    find_event_mapper_by_name,
    find_event_mapper_by_type,
)
from aioddd.testing import AsyncMock, mock


def test_event_and_event_mapper() -> None:
    class _EventTest(Event):
        @dataclass
        class Attributes:
            __slots__ = ('foo',)
            foo: str

        attributes: Attributes

    event = _EventTest(attributes=_EventTest.Attributes(foo='test'))

    assert Id.validate(event.meta.id)
    assert event.meta.type == 'event'
    assert event.attributes.foo == 'test'

    class _EventTestEventMapper(EventMapper):
        event_type = _EventTest
        service_name = 'test_service_name'
        event_name = 'test_name'

    event_mapper = _EventTestEventMapper()

    assert event_mapper.belongs_to(event)
    assert event_mapper.service_name == 'test_service_name'
    assert event_mapper.event_name == 'test_name'

    event_encoded = event_mapper.encode(event)

    assert event_encoded['attributes'] == asdict(event.attributes)
    event_decoded = event_mapper.decode(event_encoded)

    assert event_decoded.meta == event.meta
    assert event_decoded.attributes == event.attributes


def test_find_event_mapper_by_name() -> None:
    class _TestEventMapper(EventMapper):
        service_name = 'svc'
        event_name = 'name'

    mappers = [_TestEventMapper()]
    event_mapper = find_event_mapper_by_name(name='svc.name', mappers=mappers)
    assert event_mapper is mappers[0]


def test_not_find_event_mapper_by_name() -> None:
    pytest.raises(EventMapperNotFoundError, lambda: find_event_mapper_by_name(name='svc.name', mappers=[]))


def test_find_event_mapper_by_type() -> None:
    class _EventTest(Event):
        pass

    class _TestEventMapper(EventMapper):
        event_type = _EventTest

    mappers = [_TestEventMapper()]
    event_mapper = find_event_mapper_by_type(msg=_EventTest({}), mappers=mappers)
    assert event_mapper is mappers[0]


def test_not_find_event_mapper_by_type() -> None:
    class _EventTest(Event):
        pass

    pytest.raises(EventMapperNotFoundError, lambda: find_event_mapper_by_type(msg=_EventTest({}), mappers=[]))


def test_config_event_mappers() -> None:
    class _TestEventMapper1(EventMapper):
        pass

    class _TestEventMapper2(EventMapper):
        pass

    class _TestEventMapper3(EventMapper):
        pass

    sut = ConfigEventMappers(mappers=[_TestEventMapper1()])

    assert isinstance(sut.all()[0], _TestEventMapper1)

    sut.add(mappers=_TestEventMapper2())

    assert isinstance(sut.all()[1], _TestEventMapper2)

    sut.add(mappers=[_TestEventMapper3()])

    assert isinstance(sut.all()[2], _TestEventMapper3)


@pytest.mark.asyncio
async def test_event_publishers() -> None:
    event_publisher_mock1 = mock(EventPublisher, ['publish'])
    event_publisher_mock2 = mock(EventPublisher, ['publish'])
    event_publisher_mock3 = mock(EventPublisher, ['publish'])

    event_publisher_mock1.publish.return_value = None
    event_publisher_mock2.publish.return_value = None
    event_publisher_mock3.publish.return_value = None

    publisher = EventPublishers(publishers=[event_publisher_mock1])
    publisher.add(publishers=event_publisher_mock2)
    publisher.add(publishers=[event_publisher_mock3])

    await publisher.publish(events=[mock(Event)])

    event_publisher_mock1.publish.assert_called_once()
    event_publisher_mock2.publish.assert_called_once()
    event_publisher_mock3.publish.assert_called_once()


@pytest.mark.asyncio
async def test_simple_event_bus() -> None:
    event_handler_mock1 = Mock()
    event_handler_mock2 = Mock()
    event_handler_mock3 = Mock()

    class _EventTest(Event):
        pass

    event = _EventTest()

    event_handler_mock1.subscribed_to = lambda: [_EventTest]
    event_handler_mock1.handle = AsyncMock(return_value=None)
    event_handler_mock2.subscribed_to = lambda: [_EventTest]
    event_handler_mock2.handle = AsyncMock(return_value=None)
    event_handler_mock3.subscribed_to = lambda: [_EventTest]
    event_handler_mock3.handle = AsyncMock(return_value=None)

    bus = SimpleEventBus(handlers=[event_handler_mock1])
    bus.add_handler(handler=event_handler_mock2)
    bus.add_handler(handler=[event_handler_mock3])

    await bus.notify(events=[event])

    event_handler_mock1.handle.assert_called_once()
    event_handler_mock2.handle.assert_called_once()
    event_handler_mock3.handle.assert_called_once()


@pytest.mark.asyncio
async def test_internal_event_publisher() -> None:
    event_bus_mock = mock(EventBus, ['notify'])

    class _EventTest(Event):
        pass

    event = _EventTest()

    event_bus_mock.notify.return_value = None

    publisher = InternalEventPublisher(event_bus=event_bus_mock)

    await publisher.publish(events=[event])

    event_bus_mock.notify.assert_called_once()



from logging import NOTSET, Formatter, Logger, StreamHandler, getLogger
from os import getenv
from typing import Optional, Union


def get_env(key: str, default: Optional[str] = None) -> str:  # pragma: no cover
    """Get an environment variable, return default if it is empty or doesn't exist."""
    value = getenv(key, default)
    return str(default) if not value or len(value) == 0 else value


def get_simple_logger(
    name: Optional[str] = None,
    level: Union[str, int] = NOTSET,
    fmt: str = '[%(asctime)s] - %(name)s - %(levelname)s - %(message)s',
) -> Logger:  # pragma: no cover
    logger = getLogger(name)
    logger.setLevel(level)
    handler = StreamHandler()
    handler.setLevel(level)
    formatter = Formatter(fmt)
    handler.setFormatter(formatter)
    logger.addHandler(handler)
    return logger



import pytest

from aioddd import Id, IdInvalidError


def test_id_fails_with_invalid_uuid() -> None:
    pytest.raises(IdInvalidError, lambda: Id('0'))


def test_id_is_not_valid() -> None:
    assert not Id.validate('0')


def test_id_generated_is_valid() -> None:
    id_ = Id.generate()
    assert Id.validate(id_.value())


def test_id_str_returns_value() -> None:
    id_ = Id.generate()
    assert id_.value() == id_.__str__()



import pytest

from aioddd import (
    BadRequestError,
    BaseError,
    CommandNotRegisteredError,
    ConflictError,
    DateTimeInvalidError,
    EventMapperNotFoundError,
    EventNotPublishedError,
    ForbiddenError,
    IdInvalidError,
    NotFoundError,
    QueryNotRegisteredError,
    TimestampInvalidError,
    UnauthorizedError,
    UnknownError,
)


def test_base_error_raise_a_system_exit_exception_when_it_is() -> None:
    pytest.raises(SystemExit, lambda: BaseError().with_exception(SystemExit()))


def test_base_error() -> None:
    err = BaseError(id='test_id').with_exception(Exception('test'))
    assert err.id() == 'test_id'
    assert err.code() == 'code'
    assert err.title() == 'title'
    assert err.detail() == '{}'
    assert err.meta() == {'exception': 'test', 'exception_type': "<class 'Exception'>"}
    assert (
        err.__str__()
        == '''{
  "id": "test_id",
  "code": "code",
  "title": "title",
  "detail": "{}",
  "meta": {
    "exception": "test",
    "exception_type": "<class 'Exception'>"
  }
}'''
    )


def test_base_error_create_method() -> None:
    err = BaseError().create(detail={'foo': 'test'}, meta={'test': 'foo'}, id='test_id')
    assert err.id() == 'test_id'
    assert err.code() == 'code'
    assert err.title() == 'title'
    assert err.detail() == '{"foo": "test"}'
    assert err.meta() == {"test": "foo"}


def test_not_found_error() -> None:
    err = NotFoundError()
    assert err.code() == 'not_found'
    assert err.title() == 'Not found'


def test_conflict_error() -> None:
    err = ConflictError()
    assert err.code() == 'conflict'
    assert err.title() == 'Conflict'


def test_bad_request_error() -> None:
    err = BadRequestError()
    assert err.code() == 'bad_request'
    assert err.title() == 'Bad Request'


def test_unauthorized_error() -> None:
    err = UnauthorizedError()
    assert err.code() == 'unauthorized'
    assert err.title() == 'Unauthorized'


def test_forbidden_error() -> None:
    err = ForbiddenError()
    assert err.code() == 'forbidden'
    assert err.title() == 'Forbidden'


def test_unknown_error() -> None:
    err = UnknownError()
    assert err.code() == 'unknown'
    assert err.title() == 'Unknown error'


def test_id_invalid_error() -> None:
    err = IdInvalidError()
    assert err.code() == 'id_invalid'
    assert err.title() == 'Invalid id'


def test_timestamp_invalid_error() -> None:
    err = TimestampInvalidError()
    assert err.code() == 'timestamp_invalid'
    assert err.title() == 'Invalid timestamp'


def test_datetime_invalid_error() -> None:
    err = DateTimeInvalidError()
    assert err.code() == 'datetime_invalid'
    assert err.title() == 'Invalid datetime'


def test_event_mapper_not_found_error() -> None:
    err = EventMapperNotFoundError()
    assert err.code() == 'event_mapper_not_found'
    assert err.title() == 'Event Mapper not found'


def test_event_not_published_error() -> None:
    err = EventNotPublishedError()
    assert err.code() == 'event_not_published'
    assert err.title() == 'Event not published'


def test_command_not_registered_error() -> None:
    err = CommandNotRegisteredError()
    assert err.code() == 'command_not_registered_error'
    assert err.title() == 'Command not registered'


def test_query_not_registered_error() -> None:
    err = QueryNotRegisteredError()
    assert err.code() == 'query_not_registered_error'
    assert err.title() == 'Query not registered'



import sys
from typing import Any, Dict, List, Optional, Union
from unittest.mock import MagicMock, Mock, patch

if sys.version_info.major == 3 and sys.version_info.minor >= 8:
    import unittest  # pragma: no cover

    AsyncMock = getattr(unittest.mock, 'AsyncMock')  # pragma: no cover
else:

    class AsyncMock(MagicMock):  # type: ignore
        async def __call__(self, *args, **kwargs) -> None:  # type: ignore
            return super(AsyncMock, self).__call__(*args, **kwargs)


SanitizeObject = Union[Dict[Any, Any], List[Any]]


def sanitize_objects(source: SanitizeObject, affected: SanitizeObject) -> SanitizeObject:  # pragma: no cover
    for key, value in list(affected.items()) if isinstance(affected, dict) else enumerate(affected):
        if not isinstance(affected, list) and key not in source:
            affected.pop(key)
        elif isinstance(value, (dict, list)):
            affected[key] = sanitize_objects(source[key], value)
    return affected


def mock(target: Union[str, object], attributes: Optional[List[str]] = None) -> Mock:
    target_async_mock = AsyncMock()
    if not attributes:
        patch(
            target=f'{target.__module__}.{target.__name__}' if isinstance(target, object) else target,  # type: ignore
            side_effect=target_async_mock,
        )
        return target_async_mock
    for attribute in attributes:
        attribute_async_mock = AsyncMock()
        patch.object(
            target=target,
            attribute=attribute,
            side_effect=attribute_async_mock,
        )
        target_async_mock[attribute] = attribute_async_mock
    return target_async_mock



from aioddd import AggregateRoot, Event


def test_aggregates() -> None:
    class _TestAggregateRoot(AggregateRoot):
        pass

    class _TestEvent(Event):
        pass

    agg = _TestAggregateRoot()
    evt = _TestEvent({})

    assert not len(agg.pull_aggregate_events())

    agg.record_aggregate_event(evt)
    events = agg.pull_aggregate_events()

    assert len(events) == 1
    assert not len(agg.pull_aggregate_events())



import pytest

from aioddd import Container


def test_container() -> None:
    env = 'development'
    container = Container(
        {
            'config': {'environment': env},
            'services': {},
        }
    )

    assert container.__contains__('config')
    assert container.__contains__('config.environment')
    assert container.__contains__('services')

    class _Test:
        value: str = 'foo'

    class _MyService:
        pass

    class _MyAnotherService:
        pass

    tz = 'UTC'
    svc = _Test()
    my_svc = _MyService()
    my_another_svc = _MyAnotherService()

    container.set(key='config.tz', val=tz)
    container.set(key='services.foo', val=svc)
    container.resolve([(_MyService, my_svc), my_another_svc])

    assert container.__contains__('config.tz')
    assert container.__contains__('services.foo')
    assert container.__contains__(_MyService)
    assert container.__contains__(my_svc)
    assert container.__contains__(_MyService)
    assert container.__contains__(my_another_svc)

    assert container.get(key='config.environment', typ=str) == env
    assert container.get(key='config.tz', typ=str) == tz
    assert container.get(key='services.foo', typ=_Test) is svc
    assert container.get(_MyService) is my_svc
    assert container.get(my_svc) is my_svc
    assert container.get(_MyAnotherService) is my_another_svc
    assert container.get(my_another_svc) is my_another_svc

    pytest.raises(TypeError, lambda: container.get(key='services.foo', typ=str))
    pytest.raises(KeyError, lambda: container.get(key='services.bar'))

    assert not container.__contains__('')
